import functools
import typing

from modern_di import Container
from modern_di.providers.abstract import AbstractProvider


T_co = typing.TypeVar("T_co", covariant=True)
P = typing.ParamSpec("P")


class SyncInjectedFactory(AbstractProvider[T_co]):
    __slots__ = [*AbstractProvider.BASE_SLOTS, "_factory_provider"]

    def __init__(self, factory_provider: AbstractProvider[T_co]) -> None:
        super().__init__(factory_provider.scope)
        self._factory_provider = factory_provider

    async def async_resolve(self, container: Container) -> typing.Callable[[], T_co]:  # type: ignore[override]
        return self.sync_resolve(container)

    def sync_resolve(self, container: Container) -> typing.Callable[[], T_co]:  # type: ignore[override]
        return functools.partial(self._factory_provider.sync_resolve, container)

    @property
    def cast(self) -> typing.Callable[[], T_co]:  # type: ignore[override]
        return typing.cast(typing.Callable[[], T_co], self)


class AsyncInjectedFactory(AbstractProvider[T_co]):
    __slots__ = [*AbstractProvider.BASE_SLOTS, "_factory_provider"]

    def __init__(self, factory_provider: AbstractProvider[T_co]) -> None:
        super().__init__(factory_provider.scope)
        self._factory_provider = factory_provider

    async def async_resolve(self, container: Container) -> typing.Callable[[], typing.Awaitable[T_co]]:  # type: ignore[override]
        return self.sync_resolve(container)

    def sync_resolve(self, container: Container) -> typing.Callable[[], typing.Awaitable[T_co]]:  # type: ignore[override]
        return functools.partial(self._factory_provider.async_resolve, container)

    @property
    def cast(self) -> typing.Callable[[], typing.Awaitable[T_co]]:  # type: ignore[override]
        return typing.cast(typing.Callable[[], typing.Awaitable[T_co]], self)



import abc
import enum
import itertools
import typing
import uuid

from typing_extensions import override

from modern_di import Container
from modern_di.helpers.attr_getter_helpers import get_value_from_object_by_dotted_path


T_co = typing.TypeVar("T_co", covariant=True)
R = typing.TypeVar("R")
P = typing.ParamSpec("P")


class AbstractProvider(typing.Generic[T_co], abc.ABC):
    BASE_SLOTS: typing.ClassVar = ["scope", "provider_id"]

    def __init__(self, scope: enum.IntEnum) -> None:
        self.scope = scope
        self.provider_id: typing.Final = str(uuid.uuid4())

    @abc.abstractmethod
    async def async_resolve(self, container: Container) -> T_co:
        """Resolve dependency asynchronously."""

    @abc.abstractmethod
    def sync_resolve(self, container: Container) -> T_co:
        """Resolve dependency synchronously."""

    @property
    def cast(self) -> T_co:
        return typing.cast(T_co, self)

    def _check_providers_scope(self, providers: typing.Iterable[typing.Any]) -> None:
        if any(x.scope > self.scope for x in providers if isinstance(x, AbstractProvider)):
            msg = "Scope of dependency cannot be more than scope of dependent"
            raise RuntimeError(msg)

    def __getattr__(self, attr_name: str) -> typing.Any:  # noqa: ANN401
        """Get an attribute from the resolve object.

        Args:
            attr_name: name of attribute to get.

        Returns:
            An `AttrGetter` provider that will get the attribute after resolving the current provider.

        """
        if attr_name.startswith("_"):
            msg = f"'{type(self)}' object has no attribute '{attr_name}'"
            raise AttributeError(msg)

        return AttrGetter(provider=self, attr_name=attr_name)


class AbstractOverrideProvider(AbstractProvider[T_co], abc.ABC):
    def override(self, override_object: object, container: Container) -> None:
        container.override(self.provider_id, override_object)

    def reset_override(self, container: Container) -> None:
        container.reset_override(self.provider_id)


class AbstractCreatorProvider(AbstractOverrideProvider[T_co], abc.ABC):
    BASE_SLOTS: typing.ClassVar = [*AbstractProvider.BASE_SLOTS, "_args", "_kwargs", "_creator"]

    def __init__(
        self,
        scope: enum.IntEnum,
        creator: typing.Callable[P, typing.Any],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        super().__init__(scope)
        self._check_providers_scope(itertools.chain(args, kwargs.values()))
        self._creator: typing.Final = creator
        self._args: typing.Final = args
        self._kwargs: typing.Final = kwargs

    def _sync_resolve_args(self, container: Container) -> list[typing.Any]:
        return [x.sync_resolve(container) if isinstance(x, AbstractProvider) else x for x in self._args]

    def _sync_resolve_kwargs(self, container: Container) -> dict[str, typing.Any]:
        return {k: v.sync_resolve(container) if isinstance(v, AbstractProvider) else v for k, v in self._kwargs.items()}

    def _sync_build_creator(self, container: Container) -> typing.Any:  # noqa: ANN401
        return self._creator(
            *self._sync_resolve_args(container),
            **self._sync_resolve_kwargs(container),
        )

    async def _async_resolve_args(self, container: Container) -> list[typing.Any]:
        return [await x.async_resolve(container) if isinstance(x, AbstractProvider) else x for x in self._args]

    async def _async_resolve_kwargs(self, container: Container) -> dict[str, typing.Any]:
        return {
            k: await v.async_resolve(container) if isinstance(v, AbstractProvider) else v
            for k, v in self._kwargs.items()
        }

    async def _async_build_creator(self, container: Container) -> typing.Any:  # noqa: ANN401
        return self._creator(
            *await self._async_resolve_args(container),
            **await self._async_resolve_kwargs(container),
        )


class AttrGetter(AbstractProvider[T_co]):
    """Provides an attribute after resolving the wrapped provider."""

    __slots__ = [*AbstractProvider.BASE_SLOTS, "_attrs", "_provider"]

    def __init__(self, provider: AbstractProvider[T_co], attr_name: str) -> None:
        """Create a new AttrGetter instance.

        Args:
            provider: provider to wrap.
            attr_name: attribute name to resolve when the provider is resolved.

        """
        super().__init__(scope=provider.scope)
        self._provider = provider
        self._attrs = [attr_name]

    def __getattr__(self, attr: str) -> "AttrGetter[T_co]":
        if attr.startswith("_"):
            msg = f"'{type(self)}' object has no attribute '{attr}'"
            raise AttributeError(msg)
        self._attrs.append(attr)
        return self

    @override
    async def async_resolve(self, container: Container) -> typing.Any:
        resolved_provider_object = await self._provider.async_resolve(container)
        attribute_path = ".".join(self._attrs)
        return get_value_from_object_by_dotted_path(resolved_provider_object, attribute_path)

    @override
    def sync_resolve(self, container: Container) -> typing.Any:
        resolved_provider_object = self._provider.sync_resolve(container)
        attribute_path = ".".join(self._attrs)
        return get_value_from_object_by_dotted_path(resolved_provider_object, attribute_path)



import enum
import typing

from modern_di import Container
from modern_di.providers.abstract import AbstractOverrideProvider


T_co = typing.TypeVar("T_co", covariant=True)
P = typing.ParamSpec("P")


class Object(AbstractOverrideProvider[T_co]):
    __slots__ = [*AbstractOverrideProvider.BASE_SLOTS, "_obj"]

    def __init__(self, scope: enum.IntEnum, obj: T_co) -> None:
        super().__init__(scope)
        self._obj: typing.Final = obj

    async def async_resolve(self, container: Container) -> T_co:
        return self.sync_resolve(container)

    def sync_resolve(self, container: Container) -> T_co:
        container = container.find_container(self.scope)
        if (override := container.fetch_override(self.provider_id)) is not None:
            return typing.cast(T_co, override)

        return self._obj



import typing
from operator import attrgetter


def get_value_from_object_by_dotted_path(obj: typing.Any, path: str) -> typing.Any:  # noqa: ANN401
    attribute_getter = attrgetter(path)
    return attribute_getter(obj)



import contextlib
import enum
import types
import typing

from modern_di.provider_state import ProviderState


if typing.TYPE_CHECKING:
    import typing_extensions


T_co = typing.TypeVar("T_co", covariant=True)


class Container(contextlib.AbstractAsyncContextManager["Container"], contextlib.AbstractContextManager["Container"]):
    __slots__ = (
        "_is_async",
        "_overrides",
        "_provider_states",
        "_use_threading_lock",
        "context",
        "parent_container",
        "scope",
    )

    def __init__(
        self,
        *,
        scope: enum.IntEnum,
        parent_container: typing.Optional["Container"] = None,
        context: dict[str, typing.Any] | None = None,
        use_threading_lock: bool = True,
    ) -> None:
        self.scope = scope
        self.parent_container = parent_container
        self.context: dict[str, typing.Any] = context or {}
        self._is_async: bool | None = None
        self._provider_states: dict[str, ProviderState[typing.Any]] = {}
        self._overrides: dict[str, typing.Any] = parent_container._overrides if parent_container else {}  # noqa: SLF001
        self._use_threading_lock = use_threading_lock

    def _exit(self) -> None:
        self._is_async = None
        self._provider_states = {}
        self._overrides = {}
        self.context = {}

    def _check_entered(self) -> None:
        if self._is_async is None:
            msg = "Enter the context first"
            raise RuntimeError(msg)

    def build_child_container(
        self, context: dict[str, typing.Any] | None = None, scope: enum.IntEnum | None = None
    ) -> "typing_extensions.Self":
        self._check_entered()
        if scope and scope <= self.scope:
            msg = "Scope of child container must be more than current scope"
            raise RuntimeError(msg)

        if not scope:
            try:
                scope = self.scope.__class__(self.scope.value + 1)
            except ValueError as exc:
                msg = f"Max scope is reached, {self.scope.name}"
                raise RuntimeError(msg) from exc

        return self.__class__(scope=scope, parent_container=self, context=context)

    def find_container(self, scope: enum.IntEnum) -> "typing_extensions.Self":
        container = self
        if container.scope < scope:
            msg = f"Scope {scope.name} is not initialized"
            raise RuntimeError(msg)

        while container.scope > scope and container.parent_container:
            container = typing.cast("typing_extensions.Self", container.parent_container)

        if container.scope != scope:
            msg = f"Scope {scope.name} is skipped"
            raise RuntimeError(msg)

        return container

    def fetch_provider_state(
        self,
        provider_id: str,
        is_async_resource: bool = False,
        use_asyncio_lock: bool = False,
        use_threading_lock: bool = False,
    ) -> ProviderState[typing.Any]:
        self._check_entered()
        if is_async_resource and self._is_async is False:
            msg = "Resolving async resource in sync container is not allowed"
            raise RuntimeError(msg)

        if provider_state := self._provider_states.get(provider_id):
            return provider_state

        # expected to be thread-safe, because setdefault is atomic
        return self._provider_states.setdefault(
            provider_id,
            ProviderState(
                use_asyncio_lock=use_asyncio_lock,
                use_threading_lock=self._use_threading_lock and use_threading_lock,
            ),
        )

    def override(self, provider_id: str, override_object: object) -> None:
        self._overrides[provider_id] = override_object

    def fetch_override(self, provider_id: str) -> object | None:
        return self._overrides.get(provider_id)

    def reset_override(self, provider_id: str | None = None) -> None:
        if provider_id is None:
            self._overrides = {}
        else:
            self._overrides.pop(provider_id, None)

    def async_enter(self) -> "Container":
        self._is_async = True
        return self

    def sync_enter(self) -> "Container":
        self._is_async = False
        return self

    async def async_close(self) -> None:
        self._check_entered()
        for provider_state in reversed(self._provider_states.values()):
            await provider_state.async_tear_down()
        self._exit()

    def sync_close(self) -> None:
        self._check_entered()
        for provider_state in reversed(self._provider_states.values()):
            provider_state.sync_tear_down()
        self._exit()

    async def __aenter__(self) -> "Container":
        return self.async_enter()

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        traceback: types.TracebackType | None,
    ) -> None:
        await self.async_close()

    def __enter__(self) -> "Container":
        return self.sync_enter()

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: types.TracebackType | None,
    ) -> None:
        self.sync_close()
